/**************************************************
* Implements a Modbus TCP Slave (Server)
* Michael Franklin
* 2012-03-23 - First Release
* 2012-12-21 - Fix to function codes 1 and 2; would
               only return first byte.  All others
			   were 0.
**************************************************/

#include "moacon500.h"

#define     MODBUS_TCP_PROTOCOL_ID                  0x00
#define     MODBUS_TCP_EXCEPTION_WEIGHT             0x80
#define     MODBUS_TCP_EXCEPTION_ILLEGAL_FUNCTION   0x01
#define     MODBUS_TCP_WRITE_COIL_ON                0xFF00
#define     MODBUS_TCP_WRITE_COIL_OFF               0x0000

#define		MODBUS_TCP_FUNC_CODE_READ_COIL 			 1
#define    	MODBUS_TCP_FUNC_CODE_READ_INPUT 		 2
#define    	MODBUS_TCP_FUNC_CODE_READ_HOLD_REG 		 3
#define    	MODBUS_TCP_FUNC_CODE_READ_INPUT_REG 	 4
#define    	MODBUS_TCP_FUNC_CODE_WRITE_ONE_COIL 	 5
#define    	MODBUS_TCP_FUNC_CODE_WRITE_MULT_COIL 	15
#define    	MODBUS_TCP_FUNC_CODE_WRITE_ONE_REG 		 6
#define    	MODBUS_TCP_FUNC_CODE_WRITE_MULT_REG 	16

/**************************************************
* MBAP Header
**************************************************/
typedef struct
{
	u16 transactionId   __attribute__((packed));
	u16 protocolId      __attribute__((packed));
	u16 length          __attribute__((packed));
	u8 unitId;
} modbusTcpMbapHeader;

/**************************************************
* Request PDU
**************************************************/
typedef struct
{
	u8 functionCode;
	u16 startAddress    __attribute__((packed));
	u16 length          __attribute__((packed));
} modbusTcpReqPdu;

/**************************************************
* Read Response PDU
**************************************************/
typedef struct
{
	u8 functionCode;
	u8 length;
} modbusTcpRespPdu;

/**************************************************
* Write Response PDU
**************************************************/
typedef struct
{
	u8 functionCode;
	u16 startAddress    __attribute__((packed));
	u16 data            __attribute__((packed));
} modbusTcpWriteRespPdu;

/**************************************************
* Exception PDU
**************************************************/
typedef struct
{
	u8 functionCode;
	u8 exceptionCode;
} modbusTcpExceptionPdu;

/**************************************************
* Request Frame
**************************************************/
typedef struct
{
	modbusTcpMbapHeader header;
	modbusTcpReqPdu pdu;
} modbusTcpReqFrame;

/**************************************************
* Exception Frame
**************************************************/
typedef struct
{
	modbusTcpMbapHeader header;
	modbusTcpExceptionPdu pdu;
} modbusTcpExceptionFrame;

/**************************************************
* Read Response Frame
**************************************************/
typedef struct
{
	modbusTcpMbapHeader header;
	modbusTcpRespPdu pdu;
} modbusTcpRespFrame;

/**************************************************
* Write Response Frame
**************************************************/
typedef struct
{
	modbusTcpMbapHeader header;
	modbusTcpWriteRespPdu pdu;
} modbusTcpWriteRespFrame;

/**************************************************
* Gets a bit in a bit array
**************************************************/
bool modbusTcpGetBit(u8* bitArray, u16 index)
{
	return ((bitArray[index / 8] & (0x01 << (index % 8))) == (0x01 << (index % 8)));
}

/**************************************************
* Sets a bit in a bit array
**************************************************/
void modbusTcpSetBit(u8* bitArray, u16 index, bool value)
{
	bitArray[index/8] = (value ? bitArray[index/8] | (0x01 << (index % 8)) : bitArray[index/8] & ~(0x01 << (index % 8)));
}

/**************************************************
* Converts from network byte order to host byte 
* order
**************************************************/
u16 modbusTcpShortNetToHost(u16 n)
{
	return ((n & 0x00FF) << 8) | ((n & 0xFF00) >> 8);
}

/**************************************************
* Converts from host byte order to network byte 
* order
**************************************************/
u16 modbusTcpShortHostToNet(u16 n)
{
	return ((n & 0x00FF) << 8) | ((n & 0xFF00) >> 8);
}

/**************************************************
* Gets 2 bytes from the socket's receive buffer
**************************************************/
u16 modbusTcpGetShort(u8 socket)
{
	u16 value = 0;
	netRecv(socket, &value, sizeof(u16));
	return modbusTcpShortNetToHost(value);
}

/**************************************************
* Gets 2 bytes from the socket's receive buffer
**************************************************/
u8 modbusTcpGetByte(u8 socket)
{
	u8 value = 0;
	netRecv(socket, &value, sizeof(u8));
	return value;
}

/**************************************************
* Gets the MBAP header from the socket's receive
* buffer
**************************************************/
void modbusTcpGetMbapHeader(u8 socket, modbusTcpMbapHeader *header)
{
    //Get header from receive buffer
	netRecv(socket, header, sizeof(modbusTcpMbapHeader));

    //Convert from network byte order to host byte order
	header->transactionId = modbusTcpShortNetToHost(header->transactionId);
	header->protocolId = modbusTcpShortNetToHost(header->protocolId);
	header->length = modbusTcpShortNetToHost(header->length);
}

/**************************************************
* Gets a Request PDU from the socket's receive
* buffer
**************************************************/
void modbusTcpGetReqPdu(u8 socket, modbusTcpReqPdu *pdu)
{
    //Get PDU from receive buffer
	netRecv(socket, pdu, sizeof(modbusTcpReqPdu));

    //Convert from network byte order to host byte order
	pdu->startAddress = modbusTcpShortNetToHost(pdu->startAddress);
	pdu->length = modbusTcpShortNetToHost(pdu->length);
}

/**************************************************
* Sends the MBAP header on the socket
**************************************************/
void modbusTcpSendMbapHeader(u8 socket, modbusTcpMbapHeader *header)
{
    //Copy header
	modbusTcpMbapHeader sendHeader = *header;

    //Convert from host byte order to network byte order
	sendHeader.transactionId = modbusTcpShortHostToNet(sendHeader.transactionId);
	sendHeader.protocolId = modbusTcpShortHostToNet(sendHeader.protocolId);
	sendHeader.length = modbusTcpShortHostToNet(sendHeader.length);
	
    //Send
	netSend(socket, &sendHeader, sizeof(modbusTcpMbapHeader));
}

/**************************************************
* Sends an exception frame
**************************************************/
void modbusTcpSendExceptionFrame(u8 socket, modbusTcpExceptionFrame *exceptionFrame)
{
    //Send MBAP header
	modbusTcpSendMbapHeader(socket, &(exceptionFrame->header));
	
    //Copy PDU
	modbusTcpExceptionPdu pdu = exceptionFrame->pdu;	

    //Send PDU
	netSend(socket, &pdu, sizeof(modbusTcpExceptionPdu));
}

/**************************************************
* Sends a read response frame
**************************************************/
void modbusTcpSendRespFrame(u8 socket, modbusTcpRespFrame *respFrame)
{
    //Send MBAP header
	modbusTcpSendMbapHeader(socket, &(respFrame->header));
	
    //Copy Response PDU
	modbusTcpRespPdu pdu = respFrame->pdu;	

    //Send PDU
	netSend(socket, &pdu, sizeof(modbusTcpRespPdu));
}

/**************************************************
* Sends a write response frame
**************************************************/
void modbusTcpSendWriteRespFrame(u8 socket, modbusTcpWriteRespFrame *respFrame)
{
    //Send MBAP header
	modbusTcpSendMbapHeader(socket, &(respFrame->header));
	
    //Copy response pdu
	modbusTcpWriteRespPdu pdu = respFrame->pdu;

    //Convert from host byte order to network byte order
	pdu.startAddress = modbusTcpShortHostToNet(pdu.startAddress);
	pdu.data = modbusTcpShortHostToNet(pdu.data);
	
    //Send PDU
	netSend(socket, &pdu, sizeof(modbusTcpWriteRespPdu));
}

/**************************************************
* Performs a Modbus read coil, Function Code 1 or 2
**************************************************/
void modbusTcpReadCoil(u8 socket, modbusTcpReqFrame* reqFrame, u8* coilMemory)
{
    //Create response frame
	modbusTcpRespFrame respFrame;
	respFrame.header = reqFrame->header;
    respFrame.pdu.functionCode = reqFrame->pdu.functionCode;
	
    //Compute length of data
	u8 dataLength = reqFrame->pdu.length / 8;
  	dataLength += (reqFrame->pdu.length % 8) != 0 ? 1 : 0;
    
    //Compute length of packet
	respFrame.header.length = sizeof(modbusTcpRespPdu) + sizeof(u8) + dataLength;	
	respFrame.pdu.length = dataLength;
	
    //Send Response Frame
	modbusTcpSendRespFrame(socket, &respFrame);

    //Read and send data one byte at a time
	int i;
	u8 byteToSend = 0;
	for(i=0; i<reqFrame->pdu.length; i++)
	{
		bool bitValue = modbusTcpGetBit(coilMemory, reqFrame->pdu.startAddress + i);
		
		byteToSend += (bitValue << (i % 8));
		
		//if we have a full byte block
		if ((i % 8) == 7)
		{
			netSend(socket, &byteToSend, sizeof(u8));
			byteToSend = 0;
		}
	}
	
	//partial byte block needs to be sent
	if ((i % 8) != 0)
	{
		netSend(socket, &byteToSend, sizeof(u8));
	}
}

/**************************************************
* Performs a Modbus write single coil, Function 
* Code 5
**************************************************/
void modbusTcpWriteSingleCoil(u8 socket, modbusTcpReqFrame* reqFrame, u8* coilMemory)
{
    //Set the coil
	 bool value = reqFrame->pdu.length == MODBUS_TCP_WRITE_COIL_ON;
	 modbusTcpSetBit(coilMemory, reqFrame->pdu.startAddress, value);
	
    //Create response frame
	modbusTcpWriteRespFrame respFrame;
	respFrame.header = reqFrame->header;
	
	respFrame.pdu.functionCode = reqFrame->pdu.functionCode;
	respFrame.pdu.startAddress = reqFrame->pdu.startAddress;
	respFrame.pdu.data = reqFrame->pdu.length;
	
    //Send response frame
	modbusTcpSendWriteRespFrame(socket, &respFrame);
}

/**************************************************
* Performs a Modbus read registers, Function 
* Code 3 or 4
**************************************************/
void modbusTcpReadRegister(u8 socket, modbusTcpReqFrame* reqFrame, u16* registerMemory)
{
	//header
	modbusTcpRespFrame respFrame;
	respFrame.header = reqFrame->header;
	respFrame.header.length = sizeof(modbusTcpRespPdu) + sizeof(u8) + reqFrame->pdu.length * 2;	

	//pdu
	respFrame.pdu.length = reqFrame->pdu.length * 2;
	respFrame.pdu.functionCode = reqFrame->pdu.functionCode;
	modbusTcpSendRespFrame(socket, &respFrame);
	
	//data
	int i;
	u16* reg = registerMemory + reqFrame->pdu.startAddress;
	for (i=0; i<reqFrame->pdu.length; i++)
	{
		u16 regData = *(reg + i);
		
		regData = modbusTcpShortHostToNet(regData);
		netSend(socket, &regData, sizeof(u16));
	}
}

/**************************************************
* Performs a Modbus write single register, Function 
* Code 6
**************************************************/
void modbusTcpWriteSingleRegister(u8 socket, modbusTcpReqFrame* reqFrame, u16* registerMemory)
{
    //Compute address of register
	u16* reg = registerMemory + reqFrame->pdu.startAddress;

    //Write register value
	*reg = reqFrame->pdu.length;
	
    //Create response frame
	modbusTcpWriteRespFrame respFrame;
	respFrame.header = reqFrame->header;
	
	respFrame.pdu.functionCode = reqFrame->pdu.functionCode;
	respFrame.pdu.startAddress = reqFrame->pdu.startAddress;
	respFrame.pdu.data = reqFrame->pdu.length;
	
    //Send response frame
	modbusTcpSendWriteRespFrame(socket, &respFrame);
}

/**************************************************
* Performs a Modbus write multiple coils, 
* Function Code 15
**************************************************/
void modbusTcpWriteMultipleCoils(u8 socket, modbusTcpReqFrame* reqFrame, u8* coilMemory)
{
    //Compute address of coil to set
	u8 byteCount = modbusTcpGetByte(socket);

	//Set the coil
	u16 i = 0;
	u8 srcBitBlock = 0;
	for(i=0; i<reqFrame->pdu.length; i++)
	{
		//If we need a new bit block
		u8 srcBitIndex = i % 8;
		if (srcBitIndex == 0)
		{
			srcBitBlock = modbusTcpGetByte(socket);
		}
		
		bool value = modbusTcpGetBit(&srcBitBlock, srcBitIndex);
		modbusTcpSetBit(coilMemory, i, value);
	}
	
    //Create response frame
	modbusTcpWriteRespFrame respFrame;
	respFrame.header = reqFrame->header;
	respFrame.header.length = 6;

	respFrame.pdu.functionCode = reqFrame->pdu.functionCode;
	respFrame.pdu.startAddress = reqFrame->pdu.startAddress;
	respFrame.pdu.data = reqFrame->pdu.length;

    //Send response frame
	modbusTcpSendWriteRespFrame(socket, &respFrame);
}

/**************************************************
* Performs a Modbus write multiple registers, 
* Function Code 16
**************************************************/
void modbusTcpWriteMultipleRegisters(u8 socket, modbusTcpReqFrame* reqFrame, u16* registerMemory)
{
    //Compute address of first register to write
	u16* reg = registerMemory + reqFrame->pdu.startAddress;
	
    //Read, but ignore byte length from socket's receive buffer
	u8 byteLength = 0;
	netRecv(socket, &byteLength, sizeof(u8));
	
    //Write each register one at a time
	int i;
	for(i=0; i<reqFrame->pdu.length; i++)
	{
		u16* reg_i = reg + i;
		*reg_i = modbusTcpGetShort(socket);
	}
	
    //Create response frame
	modbusTcpWriteRespFrame respFrame;
	respFrame.header = reqFrame->header;
	respFrame.header.length = 6;
	
	respFrame.pdu.functionCode = reqFrame->pdu.functionCode;
	respFrame.pdu.startAddress = reqFrame->pdu.startAddress;
	respFrame.pdu.data = reqFrame->pdu.length;

    //Send response frame
	modbusTcpSendWriteRespFrame(socket, &respFrame);
}

/**************************************************
* Process request frame
**************************************************/
void modbusTcpProcessReqFrame(u8 socket, modbusTcpReqFrame* reqFrame, u8* coilMemory, u16* registerMemory)
{	
	//Discard frame if protocol id is not correct
	if (reqFrame->header.protocolId == MODBUS_TCP_PROTOCOL_ID)
	{		
		switch(reqFrame->pdu.functionCode)
		{
            //Function Code 1 and 2
			case MODBUS_TCP_FUNC_CODE_READ_COIL:
			case MODBUS_TCP_FUNC_CODE_READ_INPUT:
				modbusTcpReadCoil(socket, reqFrame, coilMemory);
				break;
				
            //Function Code 3 and 4
			case MODBUS_TCP_FUNC_CODE_READ_HOLD_REG:
			case MODBUS_TCP_FUNC_CODE_READ_INPUT_REG:	
				modbusTcpReadRegister(socket, reqFrame, registerMemory);
				break;
				
            //Function Code 5
			case MODBUS_TCP_FUNC_CODE_WRITE_ONE_COIL:
				modbusTcpWriteSingleCoil(socket, reqFrame, coilMemory);
				break;
				
            //Function Code 6
			case MODBUS_TCP_FUNC_CODE_WRITE_ONE_REG:
				modbusTcpWriteSingleRegister(socket, reqFrame, registerMemory);
				break;
            
            //Function Code 15
            case MODBUS_TCP_FUNC_CODE_WRITE_MULT_COIL:
                modbusTcpWriteMultipleCoils(socket, reqFrame, coilMemory);
                break;
				
            //Function Code 16
			case MODBUS_TCP_FUNC_CODE_WRITE_MULT_REG:
				modbusTcpWriteMultipleRegisters(socket, reqFrame, registerMemory);
				break;
				
            //Invalid Function Code, send exception
			default:
            {
					modbusTcpExceptionFrame respFrame;
					respFrame.header = reqFrame->header;
					respFrame.header.length = sizeof(modbusTcpExceptionPdu) + sizeof(u8);
					respFrame.pdu.functionCode = reqFrame->pdu.functionCode + MODBUS_TCP_EXCEPTION_WEIGHT;
					respFrame.pdu.exceptionCode = MODBUS_TCP_EXCEPTION_ILLEGAL_FUNCTION;
				
					modbusTcpSendExceptionFrame(socket, &respFrame);
					break;
            }
		}	
	}
}

void processModbusTcpRequest(u8 socket, u16* registerMemory, u8* coilMemory)
{
	static modbusTcpReqFrame frame;
	static u8 headerReceived = 0;
	
	u16 rxLength = netRxLen(socket);	
	
	//If the header is not yet received
	if (headerReceived == 0)
	{
		if (rxLength >= sizeof(modbusTcpMbapHeader))
		{
			modbusTcpGetMbapHeader(socket, &(frame.header));
			headerReceived = 1;
		}
	}
	
	//If the header is received
	if (headerReceived == 1)
	{
		rxLength = netRxLen(socket);					
		if (rxLength >= frame.header.length - 1) //-1 because we don't count the unit id
		{
			modbusTcpGetReqPdu(socket, &(frame.pdu));
			modbusTcpProcessReqFrame(socket, &frame, coilMemory, registerMemory);
			
			headerReceived = 0;
		}
	}		
}